/*
 Copyright (c) 2006-2016 Trellis Management Co. Ltd.

 This software module is an implementation of one or more tools as proposed
 for the JPEG XT standard.
 
 The copyright in this software is being made available under the
 license included with the sources LICENSE.TXT. This software may be subject to other third
 party and contributor rights, including patent rights, and no such
 rights are granted under this license. 
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <string.h>
#include "hdr_io.h"

static int rgbe_Error(rgbe_error_code error_code, const char *msg) {
	switch (error_code) {
		case rgbe_read_error:
			printf("RGBE read error\n");
			break;
		case rgbe_write_error:
			printf("RGBE write error\n");
			break;
		case rgbe_format_error:
			printf("RGBE bad file format: %s\n", msg);
			break;
		default:
		case rgbe_memory_error:
			printf("RGBE error: %s\n",msg);
	}
    
	return 0;
}

static int
rgbe_WriteMetadata(rgbeHeaderInfo *header_info) {
	header_info->gamma = 1;
	header_info->valid |= RGBE_VALID_GAMMA;
	header_info->exposure = 0;
	header_info->valid |= RGBE_VALID_EXPOSURE;
    
	return 1;
}

static void rgbe_RGBEToFloat(IMAGE_PIXELF *rgbf, unsigned char rgbe[4]) {
	if (rgbe[3]) {   // nonzero pixel
		float f = (float)(ldexp(1.0, rgbe[3] - (int)(128+8)));
		rgbf->red   = rgbe[0] * f;
		rgbf->green = rgbe[1] * f;
		rgbf->blue  = rgbe[2] * f;
        
	}
	else {
		rgbf->red = rgbf->green = rgbf->blue = 0;
	}
}

static void
rgbe_FloatToRGBE(unsigned char rgbe[4], IMAGE_PIXELF *rgbf) {
	float v;
	int e;
    
	v = rgbf->red;
	if (rgbf->green > v) v = rgbf->green;
	if (rgbf->blue > v) v = rgbf->blue;
	if (v < 1e-32) {
		rgbe[0] = rgbe[1] = rgbe[2] = rgbe[3] = 0;
	}
	else {
		v = (float)(frexp(v, &e) * 256.0 / v);
		rgbe[0] = (unsigned char) (rgbf->red * v);
		rgbe[1] = (unsigned char) (rgbf->green * v);
		rgbe[2] = (unsigned char) (rgbf->blue * v);
		rgbe[3] = (unsigned char) (e + 128);
	}
}

static int
rgbe_GetLine(FILE *file, char *buffer, int length) {
	int i;
	memset(buffer, 0, length);
	for(i = 0; i < length; i++) {
		if(!fread(&buffer[i], 1, 1, file))
			return 0;
		if(buffer[i] == 0x0A)
			break;
	}
	
	return (i < length) ? 1 : 0;
}

static int
rgbe_ReadHeader(FILE *file, unsigned *width, unsigned *height, rgbeHeaderInfo *header_info) {
	char buf[HDR_MAXLINE];
	float tempf;
	int i;
	int bFormatFound = 0;
	int bHeaderFound = 0;
    
	header_info->valid = 0;
	header_info->programtype[0] = 0;
	header_info->gamma = 1.0;
	header_info->exposure = 1.0;
    
	// get the first line
	if(!rgbe_GetLine(file, buf, HDR_MAXLINE))
		return rgbe_Error(rgbe_read_error, NULL);
    
	// check the signature
    
	if ((buf[0] != '#')||(buf[1] != '?')) {
		// if you don't want to require the magic token then comment the next line
		return rgbe_Error(rgbe_format_error,"bad initial token");
	}
	else {
		header_info->valid |= RGBE_VALID_PROGRAMTYPE;
		for(i = 0; i < sizeof(header_info->programtype) - 1; i++) {
			if((buf[i+2] == 0) || isspace(buf[i+2]))
				break;
			header_info->programtype[i] = buf[i+2];
		}
		header_info->programtype[i] = 0;
	}
    
	for(;;) {
		// get next line
		if(!rgbe_GetLine(file, buf, HDR_MAXLINE))
			return rgbe_Error(rgbe_read_error, NULL);
        
		if((buf[0] == 0) || (buf[0] == '\n')) {
			// end of header so break out of loop
			bHeaderFound = 1;
			break;
		}
		else if(strcmp(buf,"FORMAT=32-bit_rle_rgbe\n") == 0) {
			bFormatFound = 1;
		}
		else if(sscanf(buf, "GAMMA=%g", &tempf) == 1) {
			header_info->gamma = tempf;
			header_info->valid |= RGBE_VALID_GAMMA;
		}
		else if(sscanf(buf,"EXPOSURE=%g",&tempf) == 1) {
			header_info->exposure = tempf;
			header_info->valid |= RGBE_VALID_EXPOSURE;
		}
		else if((buf[0] == '#') && (buf[1] == 0x20)) {
			header_info->valid |= RGBE_VALID_COMMENT;
			strcpy(header_info->comment, buf);
		}
	}
	if(!bHeaderFound || !bFormatFound) {
		return rgbe_Error(rgbe_format_error, "invalid header");
	}
    
	// get next line
	if(!rgbe_GetLine(file , buf, HDR_MAXLINE))
		return rgbe_Error(rgbe_read_error, NULL);
    
	// get the image width & height
	if(sscanf(buf,"-Y %d +X %d", height, width) < 2) {
		if(sscanf(buf,"+X %d +Y %d", height, width) < 2) {
			return rgbe_Error(rgbe_format_error, "missing image size specifier");
		}
	}
    
	return 1;
}

static int rgbe_WriteHeader(FILE *file, unsigned width, unsigned height, rgbeHeaderInfo *info) {
	char buffer[HDR_MAXLINE];
    
	const char *programtype = "RADIANCE";
    
	if(info && (info->valid & RGBE_VALID_PROGRAMTYPE)) {
		programtype = info->programtype;
	}
	// The #? is to identify file type, the programtype is optional
	sprintf(buffer, "#?%s\n", programtype);
	if(fwrite(buffer, 1, (unsigned int)strlen(buffer), file) < 1)
		return rgbe_Error(rgbe_write_error, NULL);
	sprintf(buffer, "%s\n", info->comment);
	if(fwrite(buffer, 1, (unsigned int)strlen(buffer), file) < 1)
		return rgbe_Error(rgbe_write_error, NULL);
	sprintf(buffer, "FORMAT=32-bit_rle_rgbe\n");
	if(fwrite(buffer, 1, (unsigned int)strlen(buffer), file) < 1)
		return rgbe_Error(rgbe_write_error, NULL);
	if(info && (info->valid & RGBE_VALID_GAMMA)) {
		sprintf(buffer, "GAMMA=%g\n", info->gamma);
		if(fwrite(buffer, 1, (unsigned int)strlen(buffer), file) < 1)
			return rgbe_Error(rgbe_write_error, NULL);
	}
	if(info && (info->valid & RGBE_VALID_EXPOSURE)) {
		sprintf(buffer,"EXPOSURE=%g\n", info->exposure);
		if(fwrite(buffer, 1, (unsigned int)strlen(buffer), file) < 1)
			return rgbe_Error(rgbe_write_error, NULL);
	}
	sprintf(buffer, "\n-Y %d +X %d\n", height, width);
	if(fwrite(buffer, 1, (unsigned int)strlen(buffer), file) < 1)
		return rgbe_Error(rgbe_write_error, NULL);
    
	return 1;
}


static int
rgbe_ReadPixels(FILE *file, IMAGE_PIXELF *data, unsigned numpixels) {
    unsigned char rgbe[4];
    unsigned x;
    
    for( x = 0; x < numpixels; x++) {
        if(fread(rgbe, 1, sizeof(rgbe),file) < 1) {
            return rgbe_Error(rgbe_read_error, NULL);
        }
        rgbe_RGBEToFloat(&data[x], rgbe);
    }
    
    return 1;
}

static int
rgbe_WritePixels(FILE *file, IMAGE_PIXELF *data, unsigned numpixels) {
    unsigned char rgbe[4];
    unsigned x;
    
    for( x = 0; x < numpixels; x++) {
        rgbe_FloatToRGBE(rgbe, &data[x]);
        if(fwrite(rgbe, sizeof(rgbe), 1, file) < 1)
            return rgbe_Error(rgbe_write_error, NULL);
    }
    
    return 1;
}

static int rgbe_ReadPixels_RLE(FILE *file, IMAGE_PIXELF *data, int scanline_width, unsigned num_scanlines) {
	unsigned char rgbe[4], *scanline_buffer, *ptr, *ptr_end;
	int i, count;
	unsigned char buf[2];
	if ((scanline_width < 8)||(scanline_width > 0x7fff)) {
		// run length encoding is not allowed so read flat
		return rgbe_ReadPixels(file, data, scanline_width * num_scanlines);
	}
    scanline_buffer = NULL;
	// read in each successive scanline
	while(num_scanlines > 0) {
		if(fread(rgbe, 1, sizeof(rgbe), file) < 1) {
            free(scanline_buffer);
			return rgbe_Error(rgbe_read_error,NULL);
		}
		if((rgbe[0] != 2) || (rgbe[1] != 2) || (rgbe[2] & 0x80)) {
			// this file is not run length encoded
			rgbe_RGBEToFloat(data, rgbe);
			data ++;
            free(scanline_buffer);
			return rgbe_ReadPixels(file, data, scanline_width * num_scanlines - 1);
		}
		if((((int)rgbe[2]) << 8 | rgbe[3]) != scanline_width) {
            free(scanline_buffer);
			return rgbe_Error(rgbe_format_error,"wrong scanline width");
		}
		if(scanline_buffer == NULL) {
            scanline_buffer = (unsigned char*)malloc(sizeof(unsigned char) * 4 * scanline_width);
			if(scanline_buffer == NULL) {
				return rgbe_Error(rgbe_memory_error, "unable to allocate buffer space");
			}
		}
        
		ptr = &scanline_buffer[0];
		// read each of the four channels for the scanline into the buffer
		for(i = 0; i < 4; i++) {
			ptr_end = &scanline_buffer[(i+1)*scanline_width];
			while(ptr < ptr_end) {
				if(fread(buf, 1, 2 * sizeof(unsigned char),file) < 1) {
                    free(scanline_buffer);
					return rgbe_Error(rgbe_read_error, NULL);
				}
				if(buf[0] > 128) {
					// a run of the same value
					count = buf[0] - 128;
					if((count == 0) || (count > ptr_end - ptr)) {
                        free(scanline_buffer);
						return rgbe_Error(rgbe_format_error, "bad scanline data");
					}
					while(count-- > 0)
						*ptr++ = buf[1];
				}
				else {
					// a non-run
					count = buf[0];
					if((count == 0) || (count > ptr_end - ptr)) {
                        free(scanline_buffer);
						return rgbe_Error(rgbe_format_error, "bad scanline data");
					}
					*ptr++ = buf[1];
					if(--count > 0) {
						if(fread(ptr, 1, sizeof(unsigned char) * count, file) < 1) {
                            free(scanline_buffer);
							return rgbe_Error(rgbe_read_error, NULL);
						}
						ptr += count;
					}
				}
			}
		}

		// now convert data from buffer into floats
		for(i = 0; i < scanline_width; i++) {
            rgbe[0] = scanline_buffer[i];
			rgbe[1] = scanline_buffer[i+scanline_width];
			rgbe[2] = scanline_buffer[i+2*scanline_width];
            rgbe[3] = scanline_buffer[i+3*scanline_width];
            rgbe_RGBEToFloat(data, rgbe);

            data ++;
		}

		num_scanlines--;
	}
    
    free(scanline_buffer);
	
	return 1;
}


static int
rgbe_WriteBytes_RLE(FILE *file, unsigned char *data, int numbytes) {
	static const int MINRUNLENGTH = 4;
	int cur, beg_run, run_count, old_run_count, nonrun_count;
	unsigned char buf[2];
	
	cur = 0;
	while(cur < numbytes) {
		beg_run = cur;
		// find next run of length at least 4 if one exists
		run_count = old_run_count = 0;
		while((run_count < MINRUNLENGTH) && (beg_run < numbytes)) {
			beg_run += run_count;
			old_run_count = run_count;
			run_count = 1;
			while((data[beg_run] == data[beg_run + run_count]) && (beg_run + run_count < numbytes) && (run_count < 127))
				run_count++;
		}
		// if data before next big run is a short run then write it as such
		if ((old_run_count > 1)&&(old_run_count == beg_run - cur)) {
			buf[0] = (unsigned char)(128 + old_run_count);   // write short run
			buf[1] = data[cur];
			if(fwrite(buf, 2 * sizeof(unsigned char), 1, file) < 1)
				return rgbe_Error(rgbe_write_error, NULL);
			cur = beg_run;
		}
		// write out bytes until we reach the start of the next run
		while(cur < beg_run) {
			nonrun_count = beg_run - cur;
			if (nonrun_count > 128)
				nonrun_count = 128;
			buf[0] = (unsigned char)nonrun_count;
			if(fwrite(buf, sizeof(buf[0]), 1, file) < 1)
				return rgbe_Error(rgbe_write_error,NULL);
			if(fwrite(&data[cur], sizeof(data[0]) * nonrun_count, 1, file) < 1)
				return rgbe_Error(rgbe_write_error,NULL);
			cur += nonrun_count;
		}
		// write out next run if one was found
		if (run_count >= MINRUNLENGTH) {
			buf[0] = (unsigned char)(128 + run_count);
			buf[1] = data[beg_run];
			if(fwrite(buf, sizeof(buf[0]) * 2, 1, file) < 1)
				return rgbe_Error(rgbe_write_error,NULL);
			cur += run_count;
		}
	}
	
	return 1;
}

static int
rgbe_WritePixels_RLE(FILE *file, IMAGE_PIXELF *data, unsigned scanline_width, unsigned num_scanlines) {
	unsigned char rgbe[4];
	unsigned char *buffer;
	unsigned x;
    int i;
    
	if ((scanline_width < 8)||(scanline_width > 0x7fff)) {
		// run length encoding is not allowed so write flat
		return rgbe_WritePixels(file, data, scanline_width * num_scanlines);
	}
	buffer = (unsigned char*)malloc(sizeof(unsigned char) * 4 * scanline_width);
	if (buffer == NULL) {
		// no buffer space so write flat
		return rgbe_WritePixels(file, data, scanline_width * num_scanlines);
	}
	while(num_scanlines-- > 0) {
		rgbe[0] = (unsigned char)2;
		rgbe[1] = (unsigned char)2;
		rgbe[2] = (unsigned char)(scanline_width >> 8);
		rgbe[3] = (unsigned char)(scanline_width & 0xFF);
		if(fwrite(rgbe, sizeof(rgbe), 1, file) < 1) {
			free(buffer);
			return rgbe_Error(rgbe_write_error, NULL);
		}
		for( x = 0; x < scanline_width; x++) {
			rgbe_FloatToRGBE(rgbe, data);
			buffer[x] = rgbe[0];
			buffer[x+scanline_width] = rgbe[1];
			buffer[x+2*scanline_width] = rgbe[2];
			buffer[x+3*scanline_width] = rgbe[3];
			data ++;
		}
		// write out each of the four channels separately run length encoded
		// first red, then green, then blue, then exponent
		for( i = 0; i < 4; i++) {
			int bOK = rgbe_WriteBytes_RLE(file, &buffer[i*scanline_width], scanline_width);
			if(!bOK) {
				free(buffer);
				return bOK;
			}
		}
	}
	free(buffer);
	
	return 1;
}

FILE_IMAGE *hdr_load( const char *filename )
{
    FILE_IMAGE *image=NULL;
    unsigned width, height;
    rgbeHeaderInfo header_info;
    unsigned y;
    int row_stride;
    IMAGE_PIXELF *scanline;

    FILE *file = fopen(filename,"rb");
    if (file == NULL)
        return NULL;
    
    // Read the header
    if(rgbe_ReadHeader(file, &width, &height, &header_info) == 0) {
        fclose(file);
        return NULL;
    }
    
    row_stride = (width *3) * sizeof(float);
    
    image = (FILE_IMAGE *) malloc ( sizeof(FILE_IMAGE));
    if (image == NULL)
    {
        fclose(file);
        return NULL;
    }
    
    image->data = (unsigned char *) malloc( (width * height * 3) * sizeof(float));
    if (image->data == NULL)
    {
        fclose(file);
        free (image);
        return NULL;
    }
    scanline =(IMAGE_PIXELF *) malloc( row_stride );
    if (scanline == NULL)
    {
        fclose(file);
        return NULL;
    }
    for( y = 0; y < height; y++) {
        if(!rgbe_ReadPixels_RLE(file, scanline, width, 1)) {
            free(scanline);
            fclose(file);
            return NULL;
        }

        memcpy(image->data+(((width*3)*sizeof(float))*y), scanline, (width*3)*sizeof(float));
    }
    image->width = width;
    image->height = height;
    image->type = IMAGE_TYPE_HDR;
    image->bpp=96;
    
    image->data_size=image->width*image->height*(3*sizeof(float));
    
    free(scanline);
    fclose(file);
    return image;
}

int hdr_write( FILE_IMAGE *image, const char *filename)
{
	FILE *file;
    rgbeHeaderInfo header_info;
    unsigned width  = image->width;
	unsigned height = image->height;
    unsigned y;
    int row_stride;
    if (image->type != IMAGE_TYPE_HDR)
        return 0;
    
    file = fopen( filename, "wb");
    if (file == NULL)
    {
        return 0;
    }
    memset(&header_info, 0, sizeof(rgbeHeaderInfo));
    rgbe_WriteMetadata(&header_info);
    sprintf(header_info.comment, "# Made with XDepth functions");
    if(!rgbe_WriteHeader(file, width, height, &header_info)) {
        fclose(file);
		return 0;
	}

    row_stride = (width * 3)* sizeof(float);
    
	for( y = 0; y < height; y++) {
        IMAGE_PIXELF *scanline = (IMAGE_PIXELF*)(image->data+(row_stride*y));
		if(!rgbe_WritePixels_RLE(file, scanline, width, 1)) {
            fclose(file);
			return 0;
		}
	}
    fclose(file);
    return 1;
}
