/*
 Copyright (c) 2006-2016 Trellis Management Co. Ltd.

 This software module is an implementation of one or more tools as proposed
 for the JPEG XT standard.
 
 The copyright in this software is being made available under the
 license included with the sources LICENSE.TXT. This software may be subject to other third
 party and contributor rights, including patent rights, and no such
 rights are granted under this license. 
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "common.h"
#include "image.h"


extern int g_Rw;
extern int g_e;

// Clamp value between maxf and minf
float clamp(float value,float maxf,float minf)
{
    return value>maxf?maxf:value<minf?minf:value;
}

// String to lower
char *__strlwr(char *str)
{
    char *s=str;
    while (*s) {
        *s=tolower(*s);
        s++;
    }
    return str;
}

// Get filename extension
char *get_extension(const char *filename)
{
    char *found=NULL;
    char *p=(char *)filename;
    char *last_good=NULL;
    do {
        last_good=found;
        found=strstr(p,".");
        if (found!=NULL)
            p=found+1;
    } while(found!=NULL);
    return __strlwr(last_good+1);
}

// Free and FILE_IMAGE structure
void free_image(FILE_IMAGE *image)
{
    if (image != NULL)
    {
        if (image->data != NULL)
            free(image->data);
        free (image);
    }
}

// Free a MEMORY_FILE_IMAGE structure
void free_memory_image ( MEMORY_FILE_IMAGE *image )
{
    if (image != NULL)
    {
        if (image->data != NULL)
            free(image->data);
        free (image);
    }
}

// Linearize an input floating point image from a specific gamma value
int image_linearize(FILE_IMAGE *image, float gamma)
{
    unsigned int x,y;
    unsigned char *pimage=image->data;
    unsigned channels=3;
    unsigned img_pitch;

    if (image->bpp!=96)
        return -1;
    if (image->grayscale==1)
        channels=1;
    img_pitch=(image->width*channels)*sizeof(float);
    for (y=0; y<image->height;y++)
    {
        float *p_img_float=(float*)pimage;
        int off=0;
        for (x=0; x<image->width;x++)
        {
            p_img_float[off]=(float)pow(p_img_float[off],gamma);
            if (channels==3)
            {
                p_img_float[off+1]=(float)pow(p_img_float[off+1],gamma);
                p_img_float[off+2]=(float)pow(p_img_float[off+2],gamma);
            }
            off+=channels;
        }
        pimage+=img_pitch;
    }
    return 0;
}

// Linearize a floating point image from srgb profile
int linear_to_srgb(FILE_IMAGE *image)
{
    unsigned int x,y;
    float r,g,b;
    float p1,p2,p3;
    float one_by_p2;
    float one_add_p3;
    float p1_add_p3;
    float yc;
    unsigned channels=3;
    unsigned char *pimage=image->data;
    unsigned img_pitch;
    if (image->bpp!=96)
        return -1;
    if (image->grayscale==1)
        channels=1;
    img_pitch=(image->width*channels)*sizeof(float);
    p1=DEFAULT_P1;
    p2=DEFAULT_P2;
    p3=DEFAULT_P3;
    yc = powf(((p1 + p3)/(1.0f  + p3)),p2);
    one_by_p2=1.0f/p2;
    one_add_p3=1.0f+p3;
    p1_add_p3=p1+p3;
    for (y=0; y<image->height;y++)
    {
        float *p_img_float=(float*)pimage;
        int off=0;
        for (x=0; x<image->width;x++)
        {
            r=p_img_float[off];
            p_img_float[off]=r>=yc?powf(r,one_by_p2) * one_add_p3 - p3:(r*p1) / powf(p1_add_p3/one_add_p3,p2);
            if (channels==3)
            {
                g=p_img_float[off+1];
                b=p_img_float[off+2];
                p_img_float[off+1]=g>=yc?powf(g,one_by_p2) * one_add_p3 - p3:(g*p1) / powf(p1_add_p3/one_add_p3,p2);
                p_img_float[off+2]=b>=yc?powf(b,one_by_p2) * one_add_p3 - p3:(b*p1) / powf(p1_add_p3/one_add_p3,p2);
            }
            off+=channels;
        }
        pimage+=img_pitch;
    }
    return 0;
}

// linearize image using srgb profile
int srgb_to_linear(FILE_IMAGE *image)
{
    unsigned int x,y;
    float r,g,b;
    float p1,p2,p3;
    float one_add_p3;
    float p1_add_p3;
    unsigned channels=3;
    unsigned char *pimage=image->data;
    unsigned img_pitch;
    if (image->bpp!=96)
        return -1;
    if (image->grayscale==1)
        channels=1;
    img_pitch=(image->width*channels)*sizeof(float);
    p1=DEFAULT_P1;
    p2=DEFAULT_P2;
    p3=DEFAULT_P3             ;
    one_add_p3=1.0f+p3;
    p1_add_p3=p1+p3;
    for (y=0; y<image->height;y++)
    {
        float *p_img_float=(float*)pimage;
        int off=0;
        for (x=0; x<image->width;x++)
        {
            r=p_img_float[off];
            p_img_float[off]  = r>=p1?powf((r+p3)/one_add_p3,p2):powf(p1_add_p3 / one_add_p3,p2) * r/p1;
            if (channels==3)
            {
                g=p_img_float[off+1];
                b=p_img_float[off+2];
                p_img_float[off+1]= g>=p1?powf((g+p3)/one_add_p3,p2):powf(p1_add_p3 / one_add_p3,p2) * g/p1;
                p_img_float[off+2]= b>=p1?powf((b+p3)/one_add_p3,p2):powf(p1_add_p3 / one_add_p3,p2) * b/p1;
            }
            off+=channels;

        }
        pimage+=img_pitch;
    }
    return 0;
}

// Convert image from XYZ to RGB color space using specified matrix
FILE_IMAGE *image_convert_xyz_rgb(FILE_IMAGE *src,float matrix[])
{
    FILE_IMAGE *dest;
    unsigned int yy,xx;
    unsigned char *dest_img;
    unsigned char *src_img=src->data;
    unsigned flt_pitch;
    unsigned input_channels=3;

    if (src==NULL)
        return NULL;
    if (src->grayscale==1)
        input_channels=1;

    dest=allocate_image(src->width,src->height,96,input_channels);
    if (dest==NULL)
        return NULL;
    dest_img=dest->data;
    flt_pitch=(dest->width*3)*sizeof(float);
    for (yy=0; yy<src->height;yy++)
    {
        COLOR_VECTOR *p_dest_img=(COLOR_VECTOR*)dest_img;
        COLOR_VECTOR *p_src_img=(COLOR_VECTOR*)src_img;
        for (xx=0; xx<src->width;xx++)
        {
            float R,G,B;

            R = (matrix[0] * p_src_img[xx].x ) + (matrix[1] * p_src_img[xx].y) + (matrix[2] * p_src_img[xx].z);
            G = (matrix[3] * p_src_img[xx].x ) + (matrix[4] * p_src_img[xx].y) + (matrix[5] * p_src_img[xx].z);
            B = (matrix[6] * p_src_img[xx].x ) + (matrix[7] * p_src_img[xx].y) + (matrix[8] * p_src_img[xx].z);

            p_dest_img[xx].x=R;
            p_dest_img[xx].y=G;
            p_dest_img[xx].z=B;

        }
        dest_img+=flt_pitch;
        src_img+=flt_pitch;
    }
    return dest;

}

// Calculate minimum/maximum image values
void image_minmax_values(FILE_IMAGE *image,float *min_r,float *min_g,float *min_b,float *max_r,float *max_g,float *max_b)
{
    unsigned char *img_float;
    unsigned int y,x;
    unsigned flt_pitch;
    if (image==NULL)
        return;
    img_float=image->data;
    flt_pitch=(image->width*3)*sizeof(float);

    *min_r=10000000.0f;
    *min_g=10000000.0f;
    *min_b=10000000.0f;

    *max_r=0.0f;
    *max_g=0.0f;
    *max_b=0.0f;

    for (y=0; y<image->height;y++)
    {
        float *p_img_float=(float*)img_float;
        int off=0;
        for (x=0; x<image->width;x++)
        {
            *min_r=p_img_float[off]< *min_r?p_img_float[off]:*min_r;
            *min_g=p_img_float[off+1]< *min_g?p_img_float[off+1]:*min_g;
            *min_b=p_img_float[off+2]< *min_b?p_img_float[off+2]:*min_b;

            *max_r=p_img_float[off]> *max_r?p_img_float[off]:*max_r;
            *max_g=p_img_float[off+1]> *max_g?p_img_float[off+1]:*max_g;
            *max_b=p_img_float[off+2]> *max_b?p_img_float[off+2]:*max_b;

            off+=3;
        }
        img_float+=flt_pitch;
    }
}
// Convert image from RGB to XYZ color space using specified matrix
FILE_IMAGE *image_convert_rgb_xyz(FILE_IMAGE *src,float matrix[])
{
    FILE_IMAGE *dest;
    unsigned int yy,xx;
    unsigned char *dest_img;
    unsigned char *src_img=src->data;
    unsigned flt_pitch;
    unsigned input_channels=3;
    if (src==NULL)
        return NULL;
    if (src->grayscale==1)
        input_channels=1;
    dest=allocate_image(src->width,src->height,96,input_channels);
    if (dest==NULL)
        return NULL;
    dest_img=dest->data;
    flt_pitch=(dest->width*3)*sizeof(float);
    for (yy=0; yy<src->height;yy++)
    {
        COLOR_VECTOR *p_dest_img=(COLOR_VECTOR*)dest_img;
        COLOR_VECTOR *p_src_img=(COLOR_VECTOR*)src_img;
        for (xx=0; xx<src->width;xx++)
        {
            float X,Y,Z;

            X = (matrix[0] * p_src_img[xx].x) + ( matrix[1] * p_src_img[xx].y ) +( matrix[2] * p_src_img[xx].z);
            Y = (matrix[3] * p_src_img[xx].x) + ( matrix[4] * p_src_img[xx].y ) +( matrix[5] * p_src_img[xx].z);
            Z = (matrix[6] * p_src_img[xx].x) + ( matrix[7] * p_src_img[xx].y ) +( matrix[8] * p_src_img[xx].z);

            p_dest_img[xx].x=X;
            p_dest_img[xx].y=Y;
            p_dest_img[xx].z=Z;
        }
        dest_img+=flt_pitch;
        src_img+=flt_pitch;
    }
    return dest;
}

// Convert image from uint8 to float normalized between 0 and 1
FILE_IMAGE *image_convert_float(FILE_IMAGE *src)
{
    FILE_IMAGE *dest;
    unsigned char *img_float;
    unsigned char *img_png;
    unsigned int y,x;
	unsigned flt_pitch;
	unsigned png_pitch;
	unsigned png_channels=3;
    if (src==NULL)
        return NULL;
    if (src->grayscale==1)
        png_channels=1;
    dest=allocate_image(src->width,src->height,96,png_channels);
    if (dest==NULL)
        return NULL;
    img_float=dest->data;
    img_png=src->data;
    flt_pitch=(dest->width*png_channels)*sizeof(float);
    
    
    if (src->bpp==32)
        png_channels=4;
    png_pitch=src->width*png_channels;
    for (y=0; y<src->height;y++)
    {
        float *p_img_float=(float*)img_float;
        unsigned char *p_png=img_png;
        int off=0;
        int offpng=0;
        for (x=0; x<src->width;x++)
        {
            p_img_float[off]=toFloat(p_png[offpng],g_Rw,g_e);
            if (png_channels==3)
            {
                p_img_float[off+1]=toFloat(p_png[offpng+1],g_Rw,g_e);
                p_img_float[off+2]=toFloat(p_png[offpng+2],g_Rw,g_e);
                off+=3;
            }
            else
                off++;
            offpng+=png_channels;
        }
        img_float+=flt_pitch;
        img_png+=png_pitch;
    }
    return dest;
}

// Convert image from float to uint8 normalized between 0 and 1
FILE_IMAGE *image_convert_24(FILE_IMAGE *src)
{
    FILE_IMAGE *dest;
    unsigned char *img_float;
    unsigned char *img_png;
    unsigned int y,x;
    unsigned flt_pitch;
    unsigned png_pitch;
    unsigned png_channels=3;
    if (src==NULL)
        return NULL;
    if (src->grayscale==1)
        png_channels=1;
    dest=allocate_image(src->width,src->height,24,png_channels);
    if (dest==NULL)
        return NULL;
    img_float=src->data;
    img_png=dest->data;
    flt_pitch=(src->width*png_channels)*sizeof(float);


    if (dest->bpp==32)
        png_channels=4;
    png_pitch=dest->width*png_channels;
    for (y=0; y<dest->height;y++)
    {
        float *p_img_float=(float*)img_float;
        unsigned char *p_png=img_png;
        int off=0;
        int offpng=0;
        for (x=0; x<dest->width;x++)
        {
            float r,g,b;
            r=p_img_float[off+2];
            if (r>1.0f)
                r=1.0f;
            p_png[offpng]=toByte(p_img_float[off],g_Rw,g_e);

            if (png_channels==3)
            {
                g=p_img_float[off+1];
                if (g>1.0f)
                    g=1.0f;
                b=p_img_float[off];
                if (b>1.0f)
                    b=1.0f;
                p_png[offpng+1]=toByte(p_img_float[off+1],g_Rw,g_e);
                p_png[offpng+2]=toByte(p_img_float[off+2],g_Rw,g_e);
                off+=3;
            }
            else
                off++;
            offpng+=png_channels;
        }
        img_float+=flt_pitch;
        img_png+=png_pitch;
    }
    return dest;
}

// Toggle alpha channel from image
FILE_IMAGE *toRgb(FILE_IMAGE *rgba_image)
{
    unsigned int x,y;
	unsigned char *i=rgba_image->data;
    unsigned char *t;
	FILE_IMAGE *o = NULL;
    unsigned input_channels=3;
    if (rgba_image->grayscale==1)
        input_channels=1;
    o=allocate_image(rgba_image->width,rgba_image->height,24,input_channels);
    if (o==NULL)
        return NULL;
    t=o->data;
    for (y=0;y<rgba_image->height;y++)
    {
        int off=0;
        int off2=0;
        unsigned char *to=t;
        unsigned char *ii=i;
        for (x=0;x<rgba_image->width;x++)
        {
            to[off2]=ii[off];
            to[off2+1]=ii[off+1];
            to[off2+2]=ii[off+2];
            off+=4;
            off2+=3;
        }
        t+= (rgba_image->width*3);
        i+= (rgba_image->width*4);
    }
    return o;
}

// Create a copy of a existing image
FILE_IMAGE *image_copy(FILE_IMAGE *source)
{
    FILE_IMAGE *out;
    unsigned input_channels=3;
    if (source->grayscale==1)
        input_channels=1;
    out=allocate_image(source->width,source->height,source->bpp,input_channels);
    memcpy(out->data,source->data,source->data_size);
    return out;
}
// Allocate image space
FILE_IMAGE *allocate_image ( int width, int height, int bpp, int channels)
{
    unsigned int image_size;
    FILE_IMAGE *image=(FILE_IMAGE*)malloc(sizeof(FILE_IMAGE));
    if (image == NULL)
        return NULL;
    image->width=width;
    image->height=height;
    if (bpp==24)
    {
        image->type=IMAGE_TYPE_JPEG;
        image_size=width*height*channels;
    }
    else if(bpp==32)
    {
       image->type=IMAGE_TYPE_RGB32;
       image_size=width*height*channels;
    }
    else if (bpp == 48)
    {
        image->type=IMAGE_TYPE_RGB16;
        image_size=(width*height*channels)*sizeof(unsigned short);
    }
    else if (bpp == 96)
    {
        image->type=IMAGE_TYPE_HDR;
        image_size=(width*height*channels)*sizeof(float);
    }
    else
        return NULL;
    image->bpp=bpp;
    image->data = (unsigned char *) malloc(image_size);
    if (image->data == NULL)
    {
        free (image);
        return NULL;
    }
    image->grayscale=0;
    image->data_size=image_size;
    if (channels==1)
        image->grayscale=1;
    return image;
}

// Get scanline pointer
unsigned char *get_scanline(FILE_IMAGE *image,unsigned scanline)
{
	unsigned pitch;
    unsigned channels=3;
    if (image->grayscale==1)
        channels=1;
    pitch=(image->width*channels)*((image->bpp/3)/8);
    return image->data+(scanline*pitch);
}

// Flip image horizantally
int flip_horizontal( FILE_IMAGE *image )
{
    unsigned c;
    unsigned k;
    unsigned y;
	int line;
	unsigned char *new_bits;
	unsigned char *dst_data;
	unsigned bytespp;
	unsigned char *src_data;

    if (image==NULL)
        return 0;
    if (image->data==NULL)
        return 0;
    
    if (image->bpp>24)
        return 0;
    line = image->width * 3;
    
    
    new_bits = (unsigned char*)malloc (line * sizeof(unsigned char));
    if (new_bits == NULL)
         return 0;
    dst_data = (unsigned char *) malloc ( line * image->height);
    bytespp = line / image->width;
    src_data = (unsigned char*) image->data + line +bytespp;
                                                    
                                                      
    for ( y = 0; y < image->height; y++) {
		unsigned char *dst_data;
		unsigned char *src_data;
        unsigned char *bits = image->data + (line * y);
        memcpy(new_bits, bits, line);
        
        dst_data = (unsigned char*) bits;
        src_data = (unsigned char*) (new_bits + line - bytespp);
        for( c = 0; c < image->width; c++) {
            for( k = 0; k < bytespp; k++) {
                *dst_data++ = src_data[k];
            }
            src_data -= bytespp;
        }
    }
    return 1;
}

// Flip image vertically
int flip_vertical( FILE_IMAGE *image )
{
    unsigned char *from, *mid;
    unsigned height ;
    unsigned line_s = 0;
    unsigned line_t;
    unsigned y;
    unsigned pitch;
    unsigned image_channels=3;
    if (image == NULL)
        return 0;
    if (image->data == NULL)
        return 0;
    if (image->bpp>32)
        return 0;
    if (image->bpp==32)
        image_channels=4;
    pitch = image->width * image_channels;
    height = image->height;
    mid = (unsigned char *) malloc ( pitch * sizeof(unsigned char));
    if (mid == NULL)
        return 0;
    from = image->data;
    line_t = (height-1) * pitch;
    for (y = 0; y < height/2; y++)
    {
        memcpy(mid, from + line_s, pitch);
		memcpy(from + line_s, from + line_t, pitch);
		memcpy(from + line_t, mid, pitch);
        
		line_s += pitch;
		line_t -= pitch;

    }
    free(mid);
	return 1;
}
// Calculate minimum/maximum image values
void shift_ldr_image(FILE_IMAGE *image)
{
    unsigned char *img_float;
    unsigned int y,x;
    unsigned flt_pitch;
    float min_r,min_g,min_b;
    float max_r,max_g,max_b;
	float maxmin;
    if (image==NULL)
        return;
    img_float=image->data;
    flt_pitch=(image->width*3)*sizeof(float);

    min_r=65535.0f;
    min_g=65535.0f;
    min_b=65535.0f;

    max_r=0.0f;
    max_g=0.0f;
    max_b=0.0f;

    for (y=0; y<image->height;y++)
    {
        float *p_img_float=(float*)img_float;
        int off=0;
        for (x=0; x<image->width;x++)
        {
            min_r=p_img_float[off]< min_r?p_img_float[off]:min_r;
            min_g=p_img_float[off+1]< min_g?p_img_float[off+1]:min_g;
            min_b=p_img_float[off+2]< min_b?p_img_float[off+2]:min_b;

            max_r=p_img_float[off]> max_r?p_img_float[off]:max_r;
            max_g=p_img_float[off+1]> max_g?p_img_float[off+1]:max_g;
            max_b=p_img_float[off+2]> max_b?p_img_float[off+2]:max_b;

            off+=3;
        }
        img_float+=flt_pitch;
    }
    img_float=image->data;
    maxmin=max_r -min_r;
    for (y=0; y<image->height;y++)
    {
        float *p_img_float=(float*)img_float;
        int off=0;
        for (x=0; x<image->width;x++)
        {
            p_img_float[off]=p_img_float[off]<0.001?0.0f:p_img_float[off];
            p_img_float[off+1]=p_img_float[off+1]<0.001?0.0f:p_img_float[off+1];
            p_img_float[off+2]=p_img_float[off+2]<0.001?0.0f:p_img_float[off+2];

            //p_img_float[off] = (p_img_float[off] - min_r) / maxmin;
            //p_img_float[off+1] = (p_img_float[off+1] - min_g) / maxmin;
            //p_img_float[off+2] = (p_img_float[off+2] - min_b) / maxmin;
            off+=3;
        }
        img_float+=flt_pitch;
    }
}

MEMORY_FILE_IMAGE *file_to_mem(FILE_IMAGE *image)
{
    MEMORY_FILE_IMAGE *out=(MEMORY_FILE_IMAGE*)malloc(sizeof(MEMORY_FILE_IMAGE));
    out->bpp=image->bpp;
    out->data_size=image->data_size;
    out->height=image->height;
    out->width=image->width;
    out->type=image->type;
    out->data=(unsigned char *)malloc(image->data_size);
    memcpy(out->data,image->data,image->data_size);
    out->data_size=image->data_size;
    return out;
}

FILE_IMAGE *rgbTogray(FILE_IMAGE *rgb_image)
{
    int x,y;

    unsigned char *prgb=rgb_image->data;

    unsigned rgb_pitch=rgb_image->width*3;
    unsigned gray_pitch=rgb_image->width;

    FILE_IMAGE *gray_image=allocate_image(rgb_image->width,rgb_image->height,24,1);
    unsigned char *pgray=gray_image->data;
    for (y=0; y<rgb_image->height;y++)
    {
        unsigned char *prgb_image=prgb;
        unsigned char *pgray_image=pgray;
        int off_gray=0;
        int off=0;
        for (x=0; x<rgb_image->width;x++)
        {
            pgray_image[off_gray]=(prgb_image[off]+prgb_image[off+1]+prgb_image[off+2])/3;
            off+=3;
            off_gray++;
        }
        prgb += rgb_pitch;
        pgray+=gray_pitch;
    }
    return gray_image;
}
